// Copyright (c) Microsoft Corporation
// All rights reserved. 
//
// Licensed under the Apache License, Version 2.0 (the ""License""); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 
//
// THIS CODE IS PROVIDED ON AN  *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE, MERCHANTABLITY OR NON-INFRINGEMENT. 
//
// See the Apache Version 2.0 License for specific language governing permissions and limitations under the License.

#define NTDDI_VERSION 0x06010000

#include <v8.h>
#include <string>
#include <node_object_wrap.h>
#include <ppltasks.h>
#include "CollectionsConverter.h"
#include "CollectionsWrap.h"
#include "node-async.h"
#include "NodeRtUtils.h"
#include "OpaqueWrapper.h"
#include "WrapperBase.h"

#using <Windows.WinMD>

// this undefs fixes the issues of compiling Windows.Data.Json, Windows.Storag.FileProperties, and Windows.Stroage.Search
// Some of the node header files brings windows definitions with the same names as some of the WinRT methods
#undef DocumentProperties
#undef GetObject
#undef CreateEvent
#undef FindText
#undef SendMessage

const char* REGISTRATION_TOKEN_MAP_PROPERTY_NAME = "__registrationTokenMap__";

using namespace v8;
using namespace node;
using namespace concurrency;

namespace NodeRT { namespace Windows { namespace Devices { namespace Custom { 
  v8::Handle<v8::Value> WrapKnownDeviceTypes(::Windows::Devices::Custom::KnownDeviceTypes^ wintRtInstance);
  ::Windows::Devices::Custom::KnownDeviceTypes^ UnwrapKnownDeviceTypes(Handle<Value> value);
  
  v8::Handle<v8::Value> WrapIIOControlCode(::Windows::Devices::Custom::IIOControlCode^ wintRtInstance);
  ::Windows::Devices::Custom::IIOControlCode^ UnwrapIIOControlCode(Handle<Value> value);
  
  v8::Handle<v8::Value> WrapIOControlCode(::Windows::Devices::Custom::IOControlCode^ wintRtInstance);
  ::Windows::Devices::Custom::IOControlCode^ UnwrapIOControlCode(Handle<Value> value);
  
  v8::Handle<v8::Value> WrapCustomDevice(::Windows::Devices::Custom::CustomDevice^ wintRtInstance);
  ::Windows::Devices::Custom::CustomDevice^ UnwrapCustomDevice(Handle<Value> value);
  


  static v8::Handle<v8::Value> InitIOControlAccessModeEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("IOControlAccessMode"), enumObject);

    enumObject->Set(String::NewSymbol("any"), Integer::New(0));
    enumObject->Set(String::NewSymbol("read"), Integer::New(1));
    enumObject->Set(String::NewSymbol("write"), Integer::New(2));
    enumObject->Set(String::NewSymbol("readWrite"), Integer::New(3));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitIOControlBufferingMethodEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("IOControlBufferingMethod"), enumObject);

    enumObject->Set(String::NewSymbol("buffered"), Integer::New(0));
    enumObject->Set(String::NewSymbol("directInput"), Integer::New(1));
    enumObject->Set(String::NewSymbol("directOutput"), Integer::New(2));
    enumObject->Set(String::NewSymbol("neither"), Integer::New(3));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitDeviceAccessModeEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("DeviceAccessMode"), enumObject);

    enumObject->Set(String::NewSymbol("read"), Integer::New(0));
    enumObject->Set(String::NewSymbol("write"), Integer::New(1));
    enumObject->Set(String::NewSymbol("readWrite"), Integer::New(2));

    return scope.Close(Undefined());
  }


  static v8::Handle<v8::Value> InitDeviceSharingModeEnum(const Handle<Object> exports)
  {
    HandleScope scope;
    
    Handle<Object> enumObject = Object::New();
    exports->Set(String::NewSymbol("DeviceSharingMode"), enumObject);

    enumObject->Set(String::NewSymbol("shared"), Integer::New(0));
    enumObject->Set(String::NewSymbol("exclusive"), Integer::New(1));

    return scope.Close(Undefined());
  }



  
  class KnownDeviceTypes : public WrapperBase
  {
  public:    
    static v8::Handle<v8::Value> Init(const Handle<Object> exports)
    {
      HandleScope scope;
      
      s_constructorTemplate = Persistent<FunctionTemplate>::New(FunctionTemplate::New(New));
      s_constructorTemplate->SetClassName(String::NewSymbol("KnownDeviceTypes"));
      s_constructorTemplate->InstanceTemplate()->SetInternalFieldCount(1);
      
                              
      Local<Function> constructor = s_constructorTemplate->GetFunction();

      constructor->SetAccessor(String::NewSymbol("unknown"), UnknownGetter);

      exports->Set(String::NewSymbol("KnownDeviceTypes"), constructor);
      return scope.Close(Undefined());
    }


    virtual ::Platform::Object^ GetObjectInstance() const override
    {
      return _instance;
    }

  private:
    
    KnownDeviceTypes(::Windows::Devices::Custom::KnownDeviceTypes^ instance)
    {
      _instance = instance;
    }
    
    
    static v8::Handle<v8::Value> New(const v8::Arguments& args)
    {
      HandleScope scope;

      // in case the constructor was called without the new operator
      if (!s_constructorTemplate->HasInstance(args.This()))
      {
        if (args.Length() > 0)
        {
          std::unique_ptr<Handle<Value> []> constructorArgs(new Handle<Value>[args.Length()]);

          Handle<Value> *argsPtr = constructorArgs.get();
          for (int i = 0; i < args.Length(); i++)
          {
            argsPtr[i] = args[i];
          }

          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), constructorArgs.get());
        }
        else
        {
          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), nullptr);
        }
      }
      
      ::Windows::Devices::Custom::KnownDeviceTypes^ winRtInstance;


      if (args.Length() == 1 && OpaqueWrapper::IsOpaqueWrapper(args[0]) &&
        NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::KnownDeviceTypes^>(args[0]))
      {
        try 
        {
          winRtInstance = (::Windows::Devices::Custom::KnownDeviceTypes^) NodeRT::Utils::GetObjectInstance(args[0]);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Invalid arguments, no suitable constructor found")));
        return scope.Close(Undefined());
      }

      args.This()->SetHiddenValue(String::NewSymbol("__winRtInstance__"), True());

      KnownDeviceTypes *wrapperInstance = new KnownDeviceTypes(winRtInstance);
      wrapperInstance->Wrap(args.This());

      return args.This();
    }


  




    static Handle<Value> UnknownGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;

      try 
      {
        unsigned short result = ::Windows::Devices::Custom::KnownDeviceTypes::Unknown;
        return scope.Close(Integer::New(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    

  private:
    ::Windows::Devices::Custom::KnownDeviceTypes^ _instance;
    static Persistent<FunctionTemplate> s_constructorTemplate;

    friend v8::Handle<v8::Value> WrapKnownDeviceTypes(::Windows::Devices::Custom::KnownDeviceTypes^ wintRtInstance);
    friend ::Windows::Devices::Custom::KnownDeviceTypes^ UnwrapKnownDeviceTypes(Handle<Value> value);
    friend bool IsKnownDeviceTypesWrapper(Handle<Value> value);
  };
  Persistent<FunctionTemplate> KnownDeviceTypes::s_constructorTemplate;

  v8::Handle<v8::Value> WrapKnownDeviceTypes(::Windows::Devices::Custom::KnownDeviceTypes^ winRtInstance)
  {
    HandleScope scope;

    if (winRtInstance == nullptr)
    {
      return scope.Close(Undefined());
    }

    Handle<Object> opaqueWrapper = CreateOpaqueWrapper(winRtInstance);
    Handle<Value> args[] = {opaqueWrapper};
    return scope.Close(KnownDeviceTypes::s_constructorTemplate->GetFunction()->NewInstance(_countof(args), args));
  }

  ::Windows::Devices::Custom::KnownDeviceTypes^ UnwrapKnownDeviceTypes(Handle<Value> value)
  {
     return KnownDeviceTypes::Unwrap<KnownDeviceTypes>(value.As<Object>())->_instance;
  }

  void InitKnownDeviceTypes(Handle<Object> exports)
  {
    KnownDeviceTypes::Init(exports);
  }

  class IIOControlCode : public WrapperBase
  {
  public:    
    static v8::Handle<v8::Value> Init(const Handle<Object> exports)
    {
      HandleScope scope;
      
      s_constructorTemplate = Persistent<FunctionTemplate>::New(FunctionTemplate::New(New));
      s_constructorTemplate->SetClassName(String::NewSymbol("IIOControlCode"));
      s_constructorTemplate->InstanceTemplate()->SetInternalFieldCount(1);
      
                              
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("accessMode"), AccessModeGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("bufferingMethod"), BufferingMethodGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("controlCode"), ControlCodeGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("deviceType"), DeviceTypeGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("function"), FunctionGetter);
      
      Local<Function> constructor = s_constructorTemplate->GetFunction();


      exports->Set(String::NewSymbol("IIOControlCode"), constructor);
      return scope.Close(Undefined());
    }


    virtual ::Platform::Object^ GetObjectInstance() const override
    {
      return _instance;
    }

  private:
    
    IIOControlCode(::Windows::Devices::Custom::IIOControlCode^ instance)
    {
      _instance = instance;
    }
    
    
    static v8::Handle<v8::Value> New(const v8::Arguments& args)
    {
      HandleScope scope;

      // in case the constructor was called without the new operator
      if (!s_constructorTemplate->HasInstance(args.This()))
      {
        if (args.Length() > 0)
        {
          std::unique_ptr<Handle<Value> []> constructorArgs(new Handle<Value>[args.Length()]);

          Handle<Value> *argsPtr = constructorArgs.get();
          for (int i = 0; i < args.Length(); i++)
          {
            argsPtr[i] = args[i];
          }

          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), constructorArgs.get());
        }
        else
        {
          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), nullptr);
        }
      }
      
      ::Windows::Devices::Custom::IIOControlCode^ winRtInstance;


      if (args.Length() == 1 && OpaqueWrapper::IsOpaqueWrapper(args[0]) &&
        NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(args[0]))
      {
        try 
        {
          winRtInstance = (::Windows::Devices::Custom::IIOControlCode^) NodeRT::Utils::GetObjectInstance(args[0]);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Invalid arguments, no suitable constructor found")));
        return scope.Close(Undefined());
      }

      args.This()->SetHiddenValue(String::NewSymbol("__winRtInstance__"), True());

      IIOControlCode *wrapperInstance = new IIOControlCode(winRtInstance);
      wrapperInstance->Wrap(args.This());

      return args.This();
    }


  



    static Handle<Value> AccessModeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IIOControlCode *wrapper = IIOControlCode::Unwrap<IIOControlCode>(info.This());

      try 
      {
        ::Windows::Devices::Custom::IOControlAccessMode result = wrapper->_instance->AccessMode;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> BufferingMethodGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IIOControlCode *wrapper = IIOControlCode::Unwrap<IIOControlCode>(info.This());

      try 
      {
        ::Windows::Devices::Custom::IOControlBufferingMethod result = wrapper->_instance->BufferingMethod;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> ControlCodeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IIOControlCode *wrapper = IIOControlCode::Unwrap<IIOControlCode>(info.This());

      try 
      {
        unsigned int result = wrapper->_instance->ControlCode;
        return scope.Close(Integer::New(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> DeviceTypeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IIOControlCode *wrapper = IIOControlCode::Unwrap<IIOControlCode>(info.This());

      try 
      {
        unsigned short result = wrapper->_instance->DeviceType;
        return scope.Close(Integer::New(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> FunctionGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IIOControlCode *wrapper = IIOControlCode::Unwrap<IIOControlCode>(info.This());

      try 
      {
        unsigned short result = wrapper->_instance->Function;
        return scope.Close(Integer::New(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    


  private:
    ::Windows::Devices::Custom::IIOControlCode^ _instance;
    static Persistent<FunctionTemplate> s_constructorTemplate;

    friend v8::Handle<v8::Value> WrapIIOControlCode(::Windows::Devices::Custom::IIOControlCode^ wintRtInstance);
    friend ::Windows::Devices::Custom::IIOControlCode^ UnwrapIIOControlCode(Handle<Value> value);
    friend bool IsIIOControlCodeWrapper(Handle<Value> value);
  };
  Persistent<FunctionTemplate> IIOControlCode::s_constructorTemplate;

  v8::Handle<v8::Value> WrapIIOControlCode(::Windows::Devices::Custom::IIOControlCode^ winRtInstance)
  {
    HandleScope scope;

    if (winRtInstance == nullptr)
    {
      return scope.Close(Undefined());
    }

    Handle<Object> opaqueWrapper = CreateOpaqueWrapper(winRtInstance);
    Handle<Value> args[] = {opaqueWrapper};
    return scope.Close(IIOControlCode::s_constructorTemplate->GetFunction()->NewInstance(_countof(args), args));
  }

  ::Windows::Devices::Custom::IIOControlCode^ UnwrapIIOControlCode(Handle<Value> value)
  {
     return IIOControlCode::Unwrap<IIOControlCode>(value.As<Object>())->_instance;
  }

  void InitIIOControlCode(Handle<Object> exports)
  {
    IIOControlCode::Init(exports);
  }

  class IOControlCode : public WrapperBase
  {
  public:    
    static v8::Handle<v8::Value> Init(const Handle<Object> exports)
    {
      HandleScope scope;
      
      s_constructorTemplate = Persistent<FunctionTemplate>::New(FunctionTemplate::New(New));
      s_constructorTemplate->SetClassName(String::NewSymbol("IOControlCode"));
      s_constructorTemplate->InstanceTemplate()->SetInternalFieldCount(1);
      
                              
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("accessMode"), AccessModeGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("bufferingMethod"), BufferingMethodGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("controlCode"), ControlCodeGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("deviceType"), DeviceTypeGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("function"), FunctionGetter);
      
      Local<Function> constructor = s_constructorTemplate->GetFunction();


      exports->Set(String::NewSymbol("IOControlCode"), constructor);
      return scope.Close(Undefined());
    }


    virtual ::Platform::Object^ GetObjectInstance() const override
    {
      return _instance;
    }

  private:
    
    IOControlCode(::Windows::Devices::Custom::IOControlCode^ instance)
    {
      _instance = instance;
    }
    
    
    static v8::Handle<v8::Value> New(const v8::Arguments& args)
    {
      HandleScope scope;

      // in case the constructor was called without the new operator
      if (!s_constructorTemplate->HasInstance(args.This()))
      {
        if (args.Length() > 0)
        {
          std::unique_ptr<Handle<Value> []> constructorArgs(new Handle<Value>[args.Length()]);

          Handle<Value> *argsPtr = constructorArgs.get();
          for (int i = 0; i < args.Length(); i++)
          {
            argsPtr[i] = args[i];
          }

          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), constructorArgs.get());
        }
        else
        {
          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), nullptr);
        }
      }
      
      ::Windows::Devices::Custom::IOControlCode^ winRtInstance;


      if (args.Length() == 1 && OpaqueWrapper::IsOpaqueWrapper(args[0]) &&
        NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IOControlCode^>(args[0]))
      {
        try 
        {
          winRtInstance = (::Windows::Devices::Custom::IOControlCode^) NodeRT::Utils::GetObjectInstance(args[0]);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else if (args.Length() == 4
        && args[0]->IsInt32()
        && args[1]->IsInt32()
        && args[2]->IsInt32()
        && args[3]->IsInt32())
      {
        try
        {
          unsigned short arg0 = static_cast<unsigned short>(args[0]->Int32Value());
          unsigned short arg1 = static_cast<unsigned short>(args[1]->Int32Value());
          ::Windows::Devices::Custom::IOControlAccessMode arg2 = static_cast<::Windows::Devices::Custom::IOControlAccessMode>(args[2]->Int32Value());
          ::Windows::Devices::Custom::IOControlBufferingMethod arg3 = static_cast<::Windows::Devices::Custom::IOControlBufferingMethod>(args[3]->Int32Value());
          
          winRtInstance = ref new ::Windows::Devices::Custom::IOControlCode(arg0,arg1,arg2,arg3);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Invalid arguments, no suitable constructor found")));
        return scope.Close(Undefined());
      }

      args.This()->SetHiddenValue(String::NewSymbol("__winRtInstance__"), True());

      IOControlCode *wrapperInstance = new IOControlCode(winRtInstance);
      wrapperInstance->Wrap(args.This());

      return args.This();
    }


  



    static Handle<Value> AccessModeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IOControlCode *wrapper = IOControlCode::Unwrap<IOControlCode>(info.This());

      try 
      {
        ::Windows::Devices::Custom::IOControlAccessMode result = wrapper->_instance->AccessMode;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> BufferingMethodGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IOControlCode *wrapper = IOControlCode::Unwrap<IOControlCode>(info.This());

      try 
      {
        ::Windows::Devices::Custom::IOControlBufferingMethod result = wrapper->_instance->BufferingMethod;
        return scope.Close(Integer::New(static_cast<int>(result)));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> ControlCodeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IOControlCode *wrapper = IOControlCode::Unwrap<IOControlCode>(info.This());

      try 
      {
        unsigned int result = wrapper->_instance->ControlCode;
        return scope.Close(Integer::New(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> DeviceTypeGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IOControlCode *wrapper = IOControlCode::Unwrap<IOControlCode>(info.This());

      try 
      {
        unsigned short result = wrapper->_instance->DeviceType;
        return scope.Close(Integer::New(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> FunctionGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IOControlCode^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      IOControlCode *wrapper = IOControlCode::Unwrap<IOControlCode>(info.This());

      try 
      {
        unsigned short result = wrapper->_instance->Function;
        return scope.Close(Integer::New(result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    


  private:
    ::Windows::Devices::Custom::IOControlCode^ _instance;
    static Persistent<FunctionTemplate> s_constructorTemplate;

    friend v8::Handle<v8::Value> WrapIOControlCode(::Windows::Devices::Custom::IOControlCode^ wintRtInstance);
    friend ::Windows::Devices::Custom::IOControlCode^ UnwrapIOControlCode(Handle<Value> value);
    friend bool IsIOControlCodeWrapper(Handle<Value> value);
  };
  Persistent<FunctionTemplate> IOControlCode::s_constructorTemplate;

  v8::Handle<v8::Value> WrapIOControlCode(::Windows::Devices::Custom::IOControlCode^ winRtInstance)
  {
    HandleScope scope;

    if (winRtInstance == nullptr)
    {
      return scope.Close(Undefined());
    }

    Handle<Object> opaqueWrapper = CreateOpaqueWrapper(winRtInstance);
    Handle<Value> args[] = {opaqueWrapper};
    return scope.Close(IOControlCode::s_constructorTemplate->GetFunction()->NewInstance(_countof(args), args));
  }

  ::Windows::Devices::Custom::IOControlCode^ UnwrapIOControlCode(Handle<Value> value)
  {
     return IOControlCode::Unwrap<IOControlCode>(value.As<Object>())->_instance;
  }

  void InitIOControlCode(Handle<Object> exports)
  {
    IOControlCode::Init(exports);
  }

  class CustomDevice : public WrapperBase
  {
  public:    
    static v8::Handle<v8::Value> Init(const Handle<Object> exports)
    {
      HandleScope scope;
      
      s_constructorTemplate = Persistent<FunctionTemplate>::New(FunctionTemplate::New(New));
      s_constructorTemplate->SetClassName(String::NewSymbol("CustomDevice"));
      s_constructorTemplate->InstanceTemplate()->SetInternalFieldCount(1);
      
      Handle<Value> asyncSymbol = String::NewSymbol("__winRtAsync__");
      Handle<Function> func;
                  
      func = FunctionTemplate::New(SendIOControlAsync)->GetFunction();
      func->Set(asyncSymbol, True(), PropertyAttribute::DontEnum);
      s_constructorTemplate->PrototypeTemplate()->Set(String::NewSymbol("sendIOControlAsync"), func);
      func = FunctionTemplate::New(TrySendIOControlAsync)->GetFunction();
      func->Set(asyncSymbol, True(), PropertyAttribute::DontEnum);
      s_constructorTemplate->PrototypeTemplate()->Set(String::NewSymbol("trySendIOControlAsync"), func);
      
                  
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("inputStream"), InputStreamGetter);
      s_constructorTemplate->PrototypeTemplate()->SetAccessor(String::NewSymbol("outputStream"), OutputStreamGetter);
      
      Local<Function> constructor = s_constructorTemplate->GetFunction();

      constructor->Set(String::NewSymbol("getDeviceSelector"), FunctionTemplate::New(GetDeviceSelector)->GetFunction());
      func = FunctionTemplate::New(FromIdAsync)->GetFunction();
      func->Set(asyncSymbol, True(), PropertyAttribute::DontEnum);
      constructor->Set(String::NewSymbol("fromIdAsync"), func);

      exports->Set(String::NewSymbol("CustomDevice"), constructor);
      return scope.Close(Undefined());
    }


    virtual ::Platform::Object^ GetObjectInstance() const override
    {
      return _instance;
    }

  private:
    
    CustomDevice(::Windows::Devices::Custom::CustomDevice^ instance)
    {
      _instance = instance;
    }
    
    
    static v8::Handle<v8::Value> New(const v8::Arguments& args)
    {
      HandleScope scope;

      // in case the constructor was called without the new operator
      if (!s_constructorTemplate->HasInstance(args.This()))
      {
        if (args.Length() > 0)
        {
          std::unique_ptr<Handle<Value> []> constructorArgs(new Handle<Value>[args.Length()]);

          Handle<Value> *argsPtr = constructorArgs.get();
          for (int i = 0; i < args.Length(); i++)
          {
            argsPtr[i] = args[i];
          }

          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), constructorArgs.get());
        }
        else
        {
          return s_constructorTemplate->GetFunction()->CallAsConstructor(args.Length(), nullptr);
        }
      }
      
      ::Windows::Devices::Custom::CustomDevice^ winRtInstance;


      if (args.Length() == 1 && OpaqueWrapper::IsOpaqueWrapper(args[0]) &&
        NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::CustomDevice^>(args[0]))
      {
        try 
        {
          winRtInstance = (::Windows::Devices::Custom::CustomDevice^) NodeRT::Utils::GetObjectInstance(args[0]);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Invalid arguments, no suitable constructor found")));
        return scope.Close(Undefined());
      }

      args.This()->SetHiddenValue(String::NewSymbol("__winRtInstance__"), True());

      CustomDevice *wrapperInstance = new CustomDevice(winRtInstance);
      wrapperInstance->Wrap(args.This());

      return args.This();
    }


    static Handle<Value> SendIOControlAsync(const v8::Arguments& args)
    {
      HandleScope scope;

      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::CustomDevice^>(args.This()))
      {
        return scope.Close(Undefined());
      }

      if (args.Length() == 0 || !args[args.Length() -1]->IsFunction())
      {
          ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: No callback was given")));
          return scope.Close(Undefined());
      }

      CustomDevice *wrapper = CustomDevice::Unwrap<CustomDevice>(args.This());

      ::Windows::Foundation::IAsyncOperation<unsigned int>^ op;
    

      if (args.Length() == 4
        && NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(args[0])
        && NodeRT::Utils::IsWinRtWrapperOf<::Windows::Storage::Streams::IBuffer^>(args[1])
        && NodeRT::Utils::IsWinRtWrapperOf<::Windows::Storage::Streams::IBuffer^>(args[2]))
      {
        try
        {
          ::Windows::Devices::Custom::IIOControlCode^ arg0 = UnwrapIIOControlCode(args[0]);
          ::Windows::Storage::Streams::IBuffer^ arg1 = dynamic_cast<::Windows::Storage::Streams::IBuffer^>(NodeRT::Utils::GetObjectInstance(args[1]));
          ::Windows::Storage::Streams::IBuffer^ arg2 = dynamic_cast<::Windows::Storage::Streams::IBuffer^>(NodeRT::Utils::GetObjectInstance(args[2]));
          
          op = wrapper->_instance->SendIOControlAsync(arg0,arg1,arg2);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else 
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: no suitable overload found")));
        return scope.Close(Undefined());
      }
    
      auto opTask = create_task(op);
      uv_async_t* asyncToken = NodeUtils::Async::GetAsyncToken(args[args.Length() -1].As<Function>());

      opTask.then( [asyncToken] (task<unsigned int> t) 
      {	
        try
        {
          auto result = t.get();
          NodeUtils::Async::RunCallbackOnMain(asyncToken, [result](NodeUtils::InvokeCallbackDelegate invokeCallback) {


            TryCatch tryCatch;
            Handle<Value> error; 
            Handle<Value> arg1 = Integer::New(result);
            if (tryCatch.HasCaught())
            {
              error = tryCatch.Exception()->ToObject();
            }
            else 
            {
              error = Undefined();
            }
            if (arg1.IsEmpty()) arg1 = Undefined();
            Handle<Value> args[] = {error, arg1};

            invokeCallback(_countof(args), args);
          });
        }
        catch (Platform::Exception^ exception)
        {
          NodeUtils::Async::RunCallbackOnMain(asyncToken, [exception](NodeUtils::InvokeCallbackDelegate invokeCallback) {
             
            Handle<Value> error = NodeRT::Utils::WinRtExceptionToJsError(exception);
        
            Handle<Value> args[] = {error};
            invokeCallback(_countof(args), args);
          });
        }  		
      });

      return scope.Close(Undefined());
    }
    static Handle<Value> TrySendIOControlAsync(const v8::Arguments& args)
    {
      HandleScope scope;

      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::CustomDevice^>(args.This()))
      {
        return scope.Close(Undefined());
      }

      if (args.Length() == 0 || !args[args.Length() -1]->IsFunction())
      {
          ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: No callback was given")));
          return scope.Close(Undefined());
      }

      CustomDevice *wrapper = CustomDevice::Unwrap<CustomDevice>(args.This());

      ::Windows::Foundation::IAsyncOperation<bool>^ op;
    

      if (args.Length() == 4
        && NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::IIOControlCode^>(args[0])
        && NodeRT::Utils::IsWinRtWrapperOf<::Windows::Storage::Streams::IBuffer^>(args[1])
        && NodeRT::Utils::IsWinRtWrapperOf<::Windows::Storage::Streams::IBuffer^>(args[2]))
      {
        try
        {
          ::Windows::Devices::Custom::IIOControlCode^ arg0 = UnwrapIIOControlCode(args[0]);
          ::Windows::Storage::Streams::IBuffer^ arg1 = dynamic_cast<::Windows::Storage::Streams::IBuffer^>(NodeRT::Utils::GetObjectInstance(args[1]));
          ::Windows::Storage::Streams::IBuffer^ arg2 = dynamic_cast<::Windows::Storage::Streams::IBuffer^>(NodeRT::Utils::GetObjectInstance(args[2]));
          
          op = wrapper->_instance->TrySendIOControlAsync(arg0,arg1,arg2);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else 
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: no suitable overload found")));
        return scope.Close(Undefined());
      }
    
      auto opTask = create_task(op);
      uv_async_t* asyncToken = NodeUtils::Async::GetAsyncToken(args[args.Length() -1].As<Function>());

      opTask.then( [asyncToken] (task<bool> t) 
      {	
        try
        {
          auto result = t.get();
          NodeUtils::Async::RunCallbackOnMain(asyncToken, [result](NodeUtils::InvokeCallbackDelegate invokeCallback) {


            TryCatch tryCatch;
            Handle<Value> error; 
            Handle<Value> arg1 = Boolean::New(result);
            if (tryCatch.HasCaught())
            {
              error = tryCatch.Exception()->ToObject();
            }
            else 
            {
              error = Undefined();
            }
            if (arg1.IsEmpty()) arg1 = Undefined();
            Handle<Value> args[] = {error, arg1};

            invokeCallback(_countof(args), args);
          });
        }
        catch (Platform::Exception^ exception)
        {
          NodeUtils::Async::RunCallbackOnMain(asyncToken, [exception](NodeUtils::InvokeCallbackDelegate invokeCallback) {
             
            Handle<Value> error = NodeRT::Utils::WinRtExceptionToJsError(exception);
        
            Handle<Value> args[] = {error};
            invokeCallback(_countof(args), args);
          });
        }  		
      });

      return scope.Close(Undefined());
    }
  

    static Handle<Value> FromIdAsync(const v8::Arguments& args)
    {
      HandleScope scope;

      if (args.Length() == 0 || !args[args.Length() -1]->IsFunction())
      {
          ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: No callback was given")));
          return scope.Close(Undefined());
      }

      ::Windows::Foundation::IAsyncOperation<::Windows::Devices::Custom::CustomDevice^>^ op;
      

      if (args.Length() == 4
        && args[0]->IsString()
        && args[1]->IsInt32()
        && args[2]->IsInt32())
      {
        try
        {
          Platform::String^ arg0 = ref new Platform::String(NodeRT::Utils::StringToWchar(v8::String::Value(args[0])));
          ::Windows::Devices::Custom::DeviceAccessMode arg1 = static_cast<::Windows::Devices::Custom::DeviceAccessMode>(args[1]->Int32Value());
          ::Windows::Devices::Custom::DeviceSharingMode arg2 = static_cast<::Windows::Devices::Custom::DeviceSharingMode>(args[2]->Int32Value());
          
          op = ::Windows::Devices::Custom::CustomDevice::FromIdAsync(arg0,arg1,arg2);
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else 
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: no suitable overload found")));
        return scope.Close(Undefined());
      }
    
      auto opTask = create_task(op);
      uv_async_t* asyncToken = NodeUtils::Async::GetAsyncToken(args[args.Length() -1].As<Function>());

      opTask.then( [asyncToken] (task<::Windows::Devices::Custom::CustomDevice^> t) 
      {	
        try
        {
          auto result = t.get();
          NodeUtils::Async::RunCallbackOnMain(asyncToken, [result](NodeUtils::InvokeCallbackDelegate invokeCallback) {

            
            TryCatch tryCatch;
            Handle<Value> error; 
            Handle<Value> arg1 = WrapCustomDevice(result);
            if (tryCatch.HasCaught())
            {
              error = tryCatch.Exception()->ToObject();
            }
            else 
            {
              error = Undefined();
            }
            if (arg1.IsEmpty()) arg1 = Undefined();
            Handle<Value> args[] = {error, arg1};

            invokeCallback(_countof(args), args);
          });
        }
        catch (Platform::Exception^ exception)
        {
          NodeUtils::Async::RunCallbackOnMain(asyncToken, [exception](NodeUtils::InvokeCallbackDelegate invokeCallback) {
          
            Handle<Value> error = NodeRT::Utils::WinRtExceptionToJsError(exception);
        
            Handle<Value> args[] = {error};
            invokeCallback(_countof(args), args);
          });
        }  		
      });

      return scope.Close(Undefined());
    }

    static Handle<Value> GetDeviceSelector(const v8::Arguments& args)
    {
      HandleScope scope;

      if (args.Length() == 1
        && NodeRT::Utils::IsGuid(args[0]))
      {
        try
        {
          ::Platform::Guid arg0 = NodeRT::Utils::GuidFromJs(args[0]);
          
          Platform::String^ result;
          result = ::Windows::Devices::Custom::CustomDevice::GetDeviceSelector(arg0);
          return scope.Close(NodeRT::Utils::NewString(result->Data()));
        }
        catch (Platform::Exception ^exception)
        {
          NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
          return scope.Close(Undefined());
        }
      }
      else 
      {
        ThrowException(Exception::Error(NodeRT::Utils::NewString(L"Bad arguments: no suitable overload found")));
        return scope.Close(Undefined());
      }

      return scope.Close(Undefined());
    }

    static Handle<Value> InputStreamGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::CustomDevice^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      CustomDevice *wrapper = CustomDevice::Unwrap<CustomDevice>(info.This());

      try 
      {
        ::Windows::Storage::Streams::IInputStream^ result = wrapper->_instance->InputStream;
        return scope.Close(NodeRT::Utils::CreateExternalWinRTObject("Windows.Storage.Streams", "IInputStream", result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    
    static Handle<Value> OutputStreamGetter(Local<String> property, const AccessorInfo &info)
    {
      HandleScope scope;
      
      if (!NodeRT::Utils::IsWinRtWrapperOf<::Windows::Devices::Custom::CustomDevice^>(info.This()))
      {
        return scope.Close(Undefined());
      }

      CustomDevice *wrapper = CustomDevice::Unwrap<CustomDevice>(info.This());

      try 
      {
        ::Windows::Storage::Streams::IOutputStream^ result = wrapper->_instance->OutputStream;
        return scope.Close(NodeRT::Utils::CreateExternalWinRTObject("Windows.Storage.Streams", "IOutputStream", result));
      }
      catch (Platform::Exception ^exception)
      {
        NodeRT::Utils::ThrowWinRtExceptionInJs(exception);
        return scope.Close(Undefined());
      }
    }
    


  private:
    ::Windows::Devices::Custom::CustomDevice^ _instance;
    static Persistent<FunctionTemplate> s_constructorTemplate;

    friend v8::Handle<v8::Value> WrapCustomDevice(::Windows::Devices::Custom::CustomDevice^ wintRtInstance);
    friend ::Windows::Devices::Custom::CustomDevice^ UnwrapCustomDevice(Handle<Value> value);
    friend bool IsCustomDeviceWrapper(Handle<Value> value);
  };
  Persistent<FunctionTemplate> CustomDevice::s_constructorTemplate;

  v8::Handle<v8::Value> WrapCustomDevice(::Windows::Devices::Custom::CustomDevice^ winRtInstance)
  {
    HandleScope scope;

    if (winRtInstance == nullptr)
    {
      return scope.Close(Undefined());
    }

    Handle<Object> opaqueWrapper = CreateOpaqueWrapper(winRtInstance);
    Handle<Value> args[] = {opaqueWrapper};
    return scope.Close(CustomDevice::s_constructorTemplate->GetFunction()->NewInstance(_countof(args), args));
  }

  ::Windows::Devices::Custom::CustomDevice^ UnwrapCustomDevice(Handle<Value> value)
  {
     return CustomDevice::Unwrap<CustomDevice>(value.As<Object>())->_instance;
  }

  void InitCustomDevice(Handle<Object> exports)
  {
    CustomDevice::Init(exports);
  }

} } } } 

void init(Handle<Object> exports)
{
  if (FAILED(CoInitializeEx(nullptr, COINIT_MULTITHREADED)))
  {
    ThrowException(v8::Exception::Error(NodeRT::Utils::NewString(L"error in CoInitializeEx()")));
    return;
  }
  
  NodeRT::Windows::Devices::Custom::InitIOControlAccessModeEnum(exports);
  NodeRT::Windows::Devices::Custom::InitIOControlBufferingMethodEnum(exports);
  NodeRT::Windows::Devices::Custom::InitDeviceAccessModeEnum(exports);
  NodeRT::Windows::Devices::Custom::InitDeviceSharingModeEnum(exports);
  NodeRT::Windows::Devices::Custom::InitKnownDeviceTypes(exports);
  NodeRT::Windows::Devices::Custom::InitIIOControlCode(exports);
  NodeRT::Windows::Devices::Custom::InitIOControlCode(exports);
  NodeRT::Windows::Devices::Custom::InitCustomDevice(exports);

  NodeRT::Utils::RegisterNameSpace("Windows.Devices.Custom", exports);
}


NODE_MODULE(binding, init)